from django.test import TestCase
from AkvoDjangoFormGateway.utils.validation import (
    is_date,
    is_valid_image,
    is_number,
    is_valid_string,
    is_valid_geolocation,
)


class UtilsValidationTestCase(TestCase):
    def setUp(self):
        self.data = {
            "geo": "[37.7749,-122.4194]",
            "text": "Call +6281393",
            "number": "123",
            "image_type": "image/png",
            "date": "17-5-2022",
        }

    def test_is_valid_alphanumeric(self):
        res = is_valid_string(input=self.data["text"])
        self.assertEqual(res, True)

    def test_is_invalid_alphanumeric(self):
        res = is_valid_string(input=f'{self.data["text"]}`*`')
        self.assertEqual(res, False)

    def test_is_valid_number(self):
        res = is_number(input=self.data["number"])
        self.assertEqual(res, True)

    def test_is_invalid_number(self):
        res = is_number(input=f'{self.data["number"]}O')
        self.assertEqual(res, False)

    def test_geo_string(self):
        valid = is_valid_geolocation(json_string=self.data["geo"])
        invalid1 = is_valid_geolocation(json_string="{37.7749,-122.4194}")
        invalid2 = is_valid_geolocation(json_string="['hello','world']")
        invalid3 = is_valid_geolocation(json_string="plaintext")
        self.assertEqual(valid, True)
        self.assertEqual(invalid1, False)
        self.assertEqual(invalid2, False)
        self.assertEqual(invalid3, False)

    def test_is_valid_date(self):
        res = is_date(input=self.data["date"])
        self.assertEqual(res, True)

    def test_is_invalid_date(self):
        res = is_date(input="2022-12-3")
        self.assertEqual(res, False)

    def test_is_valid_image(self):
        res = is_valid_image(image_type=self.data["image_type"])
        self.assertEqual(res, True)

    def test_is_invalid_image(self):
        res = is_valid_image(image_type="image.jpeg")
        self.assertEqual(res, False)



from django.test import TestCase, Client
from django.core.management import call_command
from AkvoDjangoFormGateway.feed import Feed
from AkvoDjangoFormGateway.models import (
    AkvoGatewayQuestion as Questions,
    AkvoGatewayAnswer as Answers,
)

client = Client()
feed = Feed()
# Tests
# consider that 8 digit phone number will
# skip to send twillio message
phone_number = "12345678"


class TwilioEndpointTestCase(TestCase):
    def setUp(self):
        call_command(
            "form_seeder", "-f", "./backend/source/forms/1.json", "-t", True
        )

    def test_request_type(self):
        # GET not allowed
        response = client.get("/api/gateway/twilio/")
        self.assertEqual(response.status_code, 405)

        # POST allowed
        response = client.post("/api/gateway/twilio/")
        self.assertEqual(response.status_code, 200)

    def test_welcome_message(self):
        json_form = {"Body": "hi", "From": f"whatsapp:+{phone_number}"}
        response = client.post("/api/gateway/twilio/", json_form)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), feed.get_list_form())

    def test_start_survey_session(self):
        reply_text = "ready#1"

        init, form_id = feed.get_init_survey_session(text=reply_text)
        json_form = {"Body": reply_text, "From": f"whatsapp:+{phone_number}"}
        response = client.post("/api/gateway/twilio/", json_form)
        self.assertEqual(response.status_code, 200)

        datapoint = feed.get_draft_datapoint(phone=phone_number)
        survey = feed.get_form(form_id=form_id, data=datapoint)
        # First question
        question = feed.get_question(form=survey)
        self.assertEqual(response.json(), f"{question.order}. {question.text}")

        # datapoint is exist
        self.assertEqual(datapoint.phone, phone_number)
        # Form id equal with datapoint
        self.assertEqual(survey.id, datapoint.form.id)

        # Answer First question
        reply_text = "text answer"
        json_form = {"Body": reply_text, "From": f"whatsapp:+{phone_number}"}
        response = client.post("/api/gateway/twilio/", json_form)
        self.assertEqual(response.status_code, 200)

        # Photo question
        question = feed.get_question(form=survey, data=datapoint)
        self.assertEqual(response.json(), f"{question.order}. {question.text}")

        # Answer photo question
        image = "http://twilio.example/image.png"
        image_type = "image/png"
        json_form = {
            "Body": "",
            "From": f"whatsapp:+{phone_number}",
            "MediaContentType0": image_type,
            "MediaUrl0": image,
        }
        response = client.post("/api/gateway/twilio/", json_form)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            feed.validate_answer(
                text="",
                question=question,
                data=datapoint,
                image_type=image_type,
            ),
            True,
        )

        # GPS question
        question = feed.get_question(form=survey, data=datapoint)

        # Answer GPS question
        lat = "-9.1161"
        lng = "10.11"
        json_form = {
            "Body": "",
            "From": f"whatsapp:+{phone_number}",
            "Latitude": lat,
            "Longitude": lng,
        }
        text = feed.get_answer_text(
            body=None,
            image_url=None,
            lat=lat,
            lng=lng,
        )
        response = client.post("/api/gateway/twilio/", json_form)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            feed.validate_answer(
                text=text,
                question=question,
                data=datapoint,
            ),
            True,
        )

        # Phone question
        question = feed.get_question(form=survey, data=datapoint)
        # Answer Phone question
        reply_text = "0829111"
        json_form = {
            "Body": reply_text,
            "From": f"whatsapp:+{phone_number}",
        }
        response = client.post("/api/gateway/twilio/", json_form)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            feed.validate_answer(
                text=reply_text,
                question=question,
                data=datapoint,
            ),
            True,
        )

        # Single option question
        question = feed.get_question(form=survey, data=datapoint)
        # Answer Single option question
        reply_text = "optiOn 1"
        json_form = {
            "Body": reply_text,
            "From": f"whatsapp:+{phone_number}",
        }
        response = client.post("/api/gateway/twilio/", json_form)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            feed.validate_answer(
                text=reply_text,
                question=question,
                data=datapoint,
            ),
            True,
        )

        # Multiple option question
        question = feed.get_question(form=survey, data=datapoint)
        # Answer Multiple option question
        reply_text = "multi 1, Multi 2"
        json_form = {
            "Body": reply_text,
            "From": f"whatsapp:+{phone_number}",
        }
        response = client.post("/api/gateway/twilio/", json_form)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            feed.validate_answer(
                text=reply_text,
                question=question,
                data=datapoint,
            ),
            True,
        )
        answer = Answers.objects.filter(
            data=datapoint, question=question
        ).first()
        self.assertEqual(answer.options, ["multi 1", "multi 2"])

        # Date question
        question = feed.get_question(form=survey, data=datapoint)
        # Answer date question
        reply_text = "15-12-1999"
        json_form = {
            "Body": reply_text,
            "From": f"whatsapp:+{phone_number}",
        }
        response = client.post("/api/gateway/twilio/", json_form)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            feed.validate_answer(
                text=reply_text,
                question=question,
                data=datapoint,
            ),
            True,
        )
        answer = Answers.objects.filter(
            data=datapoint, question=question
        ).first()
        self.assertEqual(answer.name, reply_text)

    def test_show_options(self):
        opt_question = Questions.objects.get(pk=5)
        opt_question_output = "\n- option 1\n- option 2\n"
        self.assertEqual(
            feed.show_options(question=opt_question), opt_question_output
        )

        multi_opt_q = Questions.objects.get(pk=6)
        multi_opt_output = (
            "\n- multi 1\n- multi 2\n- multi 3\n\nYou can select more than one"
            " separated by commas. *eg: op1,op2*"
        )
        self.assertEqual(
            feed.show_options(question=multi_opt_q), multi_opt_output
        )

        non_opt_q = Questions.objects.get(pk=1)
        self.assertEqual(feed.show_options(question=non_opt_q), "")

    def test_get_init_survey_session(self):
        test1 = "ready#1"
        test2 = "#1"
        test3 = "1"
        init, form_id = feed.get_init_survey_session(text=test1)
        self.assertEqual(init, True)
        init, form_id = feed.get_init_survey_session(text=test2)
        self.assertEqual(init, False)
        init, form_id = feed.get_init_survey_session(text=test3)
        self.assertEqual(init, False)

    def test_instance_request(self):
        form_id = 1

        json_form = {}
        response = client.post(
            f"/api/gateway/twilio/{form_id}/?format=json", json_form
        )
        self.assertEqual(response.status_code, 200)
        # first question shown
        fq = Questions.objects.filter(form=form_id).order_by("order").first()

        reply_text = "answer first question"
        json_form = {"Body": reply_text, "From": f"whatsapp:+{phone_number}"}

        response = client.post(f"/api/gateway/twilio/{form_id}/", json_form)
        datapoint = feed.get_draft_datapoint(phone=phone_number)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            feed.validate_answer(text=reply_text, question=fq, data=datapoint),
            True,
        )



from django.urls import path
from .views import AkvoFormViewSet, CheckView, TwilioViewSet


urlpatterns = [
    path("check/", CheckView.as_view({"get": "check"}), name="twilio"),
    path(
        "forms/",
        AkvoFormViewSet.as_view({"get": "list"}),
        name="twilio",
    ),
    path(
        "forms/<int:pk>/",
        AkvoFormViewSet.as_view({"get": "retrieve"}),
        name="twilio",
    ),
    path("twilio/", TwilioViewSet.as_view({"post": "create"}), name="twilio"),
    path(
        "twilio/<int:pk>/",
        TwilioViewSet.as_view({"post": "instance"}),
        name="twilio",
    ),
]



from django.core.management import BaseCommand
from AkvoDjangoFormGateway.models import AkvoGatewayData as FormData
from AkvoDjangoFormGateway.utils.functions import seed_data


class Command(BaseCommand):
    help = "Command to generate fake data for data & answer as test purpose"

    def add_arguments(self, parser):
        parser.add_argument(
            "-r", "--repeat", nargs="?", const=20, default=20, type=int
        )
        parser.add_argument(
            "-t", "--test", nargs="?", const=False, default=False, type=bool
        )

    def handle(self, *args, **options):
        test = options.get("test")
        repeat = options.get("repeat")
        FormData.objects.all().delete()
        seed_data(repeat=repeat, test=test)



from django.db import models
from .constants import QuestionTypes, StatusTypes


# Create your models here.
class AkvoGatewayForm(models.Model):
    name = models.CharField(max_length=255)
    description = models.TextField(null=True, default=None)
    version = models.IntegerField(default=1)

    def __str__(self):
        return self.name

    class Meta:
        db_table = "ag_form"


class AkvoGatewayQuestion(models.Model):
    form = models.ForeignKey(
        to=AkvoGatewayForm,
        on_delete=models.CASCADE,
        related_name="ag_form_questions",
    )
    order = models.BigIntegerField(null=True, default=None)
    text = models.TextField()
    type = models.IntegerField(choices=QuestionTypes.FieldStr.items())
    required = models.BooleanField(null=True, default=True)

    def __str__(self):
        return self.text

    class Meta:
        db_table = "ag_question"


class AkvoGatewayQuestionOption(models.Model):
    question = models.ForeignKey(
        to=AkvoGatewayQuestion,
        on_delete=models.CASCADE,
        related_name="ag_question_question_options",
    )
    order = models.BigIntegerField(null=True, default=None)
    code = models.CharField(max_length=255, default=None, null=True)
    name = models.TextField()

    def __str__(self):
        return self.name

    class Meta:
        db_table = "ag_option"


class AkvoGatewayData(models.Model):
    name = models.TextField()
    form = models.ForeignKey(
        to=AkvoGatewayForm,
        on_delete=models.CASCADE,
        related_name="ag_form_data",
    )
    geo = models.JSONField(null=True, default=None)
    phone = models.CharField(max_length=25)
    status = models.IntegerField(default=StatusTypes.draft)
    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(default=None, null=True)

    def __str__(self):
        return self.name

    class Meta:
        db_table = "ag_data"


class AkvoGatewayAnswer(models.Model):
    data = models.ForeignKey(
        to=AkvoGatewayData,
        on_delete=models.CASCADE,
        related_name="ag_data_answer",
    )
    question = models.ForeignKey(
        to=AkvoGatewayQuestion,
        on_delete=models.CASCADE,
        related_name="ag_question_answer",
    )
    name = models.TextField(null=True, default=None)
    value = models.FloatField(null=True, default=None)
    options = models.JSONField(default=None, null=True)

    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(default=None, null=True)

    def __str__(self):
        return self.data.name

    class Meta:
        db_table = "ag_answer"



from twilio.rest import Client
from django.conf import settings

from django.shortcuts import get_object_or_404
from rest_framework.decorators import permission_classes
from rest_framework.viewsets import (
    ModelViewSet,
    ViewSet,
    GenericViewSet,
)
from rest_framework.permissions import AllowAny
from rest_framework.response import Response
from .models import (
    AkvoGatewayForm as Forms,
    AkvoGatewayData as FormData,
)
from .serializers import ListFormSerializer, TwilioSerializer
from .constants import StatusTypes
from .feed import Feed

account_sid = settings.TWILIO_ACCOUNT_SID
auth_token = settings.TWILIO_AUTH_TOKEN
from_number = settings.TWILIO_PHONE_NUMBER
client = Client(account_sid, auth_token)


@permission_classes([AllowAny])
class CheckView(GenericViewSet):
    def check(self, request):
        return Response({"message": settings.TWILIO_ACCOUNT_SID})


class AkvoFormViewSet(ModelViewSet):
    serializer_class = ListFormSerializer
    queryset = Forms.objects.all()


class TwilioViewSet(ViewSet):
    def create(self, request, form_instance: Forms = None):
        serializer = TwilioSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        feed = Feed()

        image_type = serializer.validated_data.get("MediaContentType0")
        image_url = serializer.validated_data.get("MediaUrl0")
        phone = serializer.data.get("phone")
        body = serializer.validated_data.get("Body")
        lat = serializer.validated_data.get("Latitude")
        lng = serializer.validated_data.get("Longitude")
        text = feed.get_answer_text(
            body=body,
            image_url=image_url,
            lat=lat,
            lng=lng,
        )
        init, form_id = feed.get_init_survey_session(
            text=text, form=form_instance
        )
        datapoint = feed.get_draft_datapoint(phone=phone)
        survey = (
            form_instance
            if form_instance
            else feed.get_form(form_id=form_id, data=datapoint)
        )
        lq = feed.get_question(form=survey, data=datapoint)
        message = None
        if text in feed.welcome and not datapoint:
            message = feed.get_list_form()

        if init and not datapoint:
            dp_name = f"{survey.id}-{phone}"
            # create new survey session by creating new datapoint
            FormData.objects.create(
                form=survey,
                name=dp_name,
                phone=phone,
                status=StatusTypes.draft,
            )
            message = f"{lq.order}. {lq.text}"

        if datapoint and lq:
            valid_answer = feed.validate_answer(
                text=text, question=lq, data=datapoint, image_type=image_type
            )
            if valid_answer:
                feed.insert_answer(text=text, question=lq, data=datapoint)
                nq = feed.get_last_question(data=datapoint)
                if nq:
                    # show next question
                    message = f"{nq.order}. {nq.text}"
                    message += feed.show_options(question=nq)
                else:
                    feed.set_as_completed(data=datapoint)
                    message = "Thank you!"
            else:
                message = f"{lq.order}. {lq.text}"
                message += feed.show_options(question=lq)

        feed.send_to_client(
            client=client,
            from_number=from_number,
            body=message,
            to_number=phone,
        )
        return Response(message)

    def instance(self, request, pk=None):
        queryset = Forms.objects.all()
        form = get_object_or_404(queryset, pk=pk)

        return self.create(request=request, form_instance=form)



from django.test import TestCase
from AkvoDjangoFormGateway.utils.functions import seed_data
from AkvoDjangoFormGateway.models import (
    AkvoGatewayForm as Forms,
    AkvoGatewayQuestion as Questions,
    AkvoGatewayQuestionOption as QO,
    AkvoGatewayData as FormData,
    AkvoGatewayAnswer as Answers,
)

from AkvoDjangoFormGateway.serializers import (
    ListDataSerializer,
    ListDataAnswerSerializer,
)
from AkvoDjangoFormGateway.constants import QuestionTypes


class DataSerializerTestCase(TestCase):
    def setUp(self):
        form = Forms.objects.create(
            name="Form Test", description="Test description"
        )
        questions = [
            {"question": "Your full name", "type": QuestionTypes.text},
            {
                "question": "Check if any",
                "type": QuestionTypes.multiple_option,
                "options": ["Toilet", "Hand Washing Facilities"],
            },
            {
                "question": "Gender",
                "type": QuestionTypes.option,
                "options": ["Male", "Female"],
            },
            {
                "question": "Date",
                "type": QuestionTypes.date,
            },
            {
                "question": "Gender",
                "type": QuestionTypes.number,
            },
            {
                "question": "Geo",
                "type": QuestionTypes.geo,
            },
        ]
        for q in questions:
            question = Questions.objects.create(
                form=form, text=q["question"], type=q["type"]
            )
            if "options" in q:
                for opt in q["options"]:
                    QO.objects.create(question=question, name=opt)
        seed_data(repeat=2)
        self.data = FormData.objects.all()
        self.data_serializer = ListDataSerializer(instance=self.data.first())

        fd = self.data.first()
        self.answer = Answers.objects.filter(data=fd.id).first()
        self.answer_serializer = ListDataAnswerSerializer(instance=self.answer)

    def test_expected_fields_in_data(self):
        data = self.data_serializer.data
        self.assertCountEqual(
            list(data),
            [
                "id",
                "name",
                "form",
                "geo",
                "phone",
                "status",
                "created",
                "updated",
            ],
        )

    def test_expected_fields_in_answer(self):
        data = self.answer_serializer.data
        self.assertEqual(list(data), ["question", "value"])
        self.assertEqual(str(self.data.first()), self.data.first().name)

    def test_value_content_in_answer(self):
        data = self.answer_serializer.data
        self.assertEqual(data["value"], self.answer.name)
        self.assertEqual(str(self.answer), self.answer.data.name)



from django.test import TestCase
from AkvoDjangoFormGateway.models import AkvoGatewayForm as Forms
from AkvoDjangoFormGateway.models import AkvoGatewayQuestion as Questions
from AkvoDjangoFormGateway.models import AkvoGatewayQuestionOption as QO
from AkvoDjangoFormGateway.serializers import (
    ListQuestionSerializer,
    ListOptionSerializer,
    QuestionDefinitionSerializer,
)
from AkvoDjangoFormGateway.constants import QuestionTypes


class QuestionSerializerTestCase(TestCase):
    def setUp(self):
        form = Forms.objects.create(
            name='Test form', description='Test description'
        )
        self.question_attributes = {
            'form': form,
            'text': 'Question #1',
            'type': QuestionTypes.text,
        }

        self.question = Questions.objects.create(**self.question_attributes)
        self.serializer = ListQuestionSerializer(instance=self.question)

    def test_contains_expected_fields(self):
        data = self.serializer.data
        self.assertCountEqual(
            set(data.keys()),
            set(['id', 'form', 'order', 'text']),
        )

    def test_text_field_content(self):
        data = self.serializer.data
        self.assertEqual(data['text'], self.question_attributes['text'])
        self.assertEqual(str(self.question), self.question.text)

    def test_option_field_content(self):
        self.question_attributes['type'] = QuestionTypes.option
        for opt in ['Yes', 'No']:
            QO.objects.create(question=self.question, name=opt)
        serializer = QuestionDefinitionSerializer(
            instance=self.question,
            data=self.question_attributes,
        )
        self.assertTrue(serializer.is_valid())
        serializer.save()
        self.assertEqual('Yes' in serializer.data['options'], True)

        option = QO.objects.get(name="Yes")
        self.assertEqual(str(option), option.name)

        serializer = ListOptionSerializer(instance=option)
        self.assertEqual(serializer.data['name'], 'Yes')

    def test_type_must_be_in_choices(self):
        self.question_attributes['type'] = QuestionTypes.geo
        self.question.type = self.question_attributes['type']
        self.question.save()
        serializer = QuestionDefinitionSerializer(
            instance=self.question, data=self.question_attributes
        )
        self.assertTrue(serializer.is_valid())
        self.assertEqual(
            serializer.data['question_type'],
            QuestionTypes.FieldStr.get(QuestionTypes.geo),
        )



import os
from django.test import TestCase


class EnvironmentValueTest(TestCase):
    def test_environment_variables_present(self):
        expected_variables = [
            "TWILIO_ACCOUNT_SID",
            "TWILIO_AUTH_TOKEN",
            "TWILIO_PHONE_NUMBER",
        ]

        for variable in expected_variables:
            with self.subTest(variable=variable):
                self.assertIn(variable, os.environ)



import re
import json
from datetime import datetime


def is_number(input: str) -> bool:
    trans = str.maketrans("", "", "".join(["-", ".", ","]))
    return str(input).translate(trans).isdigit()


def is_date(input: str) -> bool:
    try:
        date_format = "%d-%m-%Y"
        datetime.strptime(input, date_format)
    except ValueError:
        return False
    return True


def is_valid_geolocation(json_string: str):
    # This line will be excluded from coverage
    # since whatsapp already validate the lat long format
    # pragma: no cover
    try:
        data = json.loads(json_string)
        if isinstance(data, list) and len(data) == 2:
            lat, lng = data
            return is_number(lat) and is_number(lng)
    except (json.JSONDecodeError, ValueError):
        pass
    return False


def is_valid_image(image_type: str) -> bool:
    return image_type in ["image/jpeg", "image/jpg", "image/png", "image/gif"]


def is_valid_string(input: str) -> bool:
    pattern = r'^[a-zA-Z0-9 !\'"\-?,.%&*()+/]+$'
    return bool(re.match(pattern, input))



import json
from django.core.management.base import BaseCommand, CommandError
from AkvoDjangoFormGateway.models import AkvoGatewayForm as Forms
from AkvoDjangoFormGateway.models import AkvoGatewayQuestion as Questions
from AkvoDjangoFormGateway.models import AkvoGatewayQuestionOption as QO
from AkvoDjangoFormGateway.constants import QuestionTypes


def seed_questions(form: Forms, questions: list):
    for qi, q in enumerate(questions):
        question = Questions.objects.filter(pk=q["id"]).first()
        if not question:
            question = Questions.objects.create(
                id=q["id"],
                form=form,
                order=qi + 1,
                text=q["question"],
                type=getattr(QuestionTypes, q["type"]),
                required=q.get("required"),
            )
        else:
            if "order" in q:
                question.order = q["order"]
            else:
                question.order = qi + 1
            question.text = q["question"]
            question.type = getattr(QuestionTypes, q["type"])
            question.required = q.get("required")
            question.save()
        if q.get("options"):
            QO.objects.filter(question=question).all().delete()
            QO.objects.bulk_create(
                [
                    QO(
                        name=o["name"].strip(),
                        question=question,
                        order=io + 1,
                    )
                    for io, o in enumerate(q.get("options"))
                ]
            )


def seed_form(
    self, data: dict, form: Forms = None, test: bool = False
) -> Forms:
    if not form:
        form = Forms.objects.create(
            id=data["id"],
            name=data["form"],
            description=data.get("description"),
            version=1,
        )
        self.stdout.write("Form Created | %s V%s" % (form.name, form.version))
    else:
        form.name = data["form"]
        form.description = data.get("description")
        form.version += 1
        form.save()
        self.stdout.write("Form Updated | %s V%s" % (form.name, form.version))
    return form


class Command(BaseCommand):
    help = "Command to insert form from json file to database"

    def add_arguments(self, parser):
        parser.add_argument(
            "-t", "--test", nargs="?", const=1, default=False, type=bool
        )
        parser.add_argument("-f", "--file", nargs="?", type=str)

    def handle(self, *args, **options):
        test = options.get("test")
        try:
            JSON_FILE = options.get("file")
            with open(f"{JSON_FILE}") as json_file:
                json_form = json.load(json_file)
            for jf in json_form:
                form = Forms.objects.filter(id=jf["id"]).first()
                form = seed_form(self, data=jf, form=form, test=test)
                if form:
                    seed_questions(form=form, questions=jf["questions"])
        except Exception as e:
            raise CommandError(e)



from io import StringIO
from django.test.utils import override_settings
from django.core.management import call_command
from django.core.management.base import CommandError
from django.test import TestCase
from AkvoDjangoFormGateway.models import AkvoGatewayForm as Forms


@override_settings(USE_TZ=False)
class FormSeederTestCase(TestCase):
    def call_form_seeder_command(self, *args, **kwargs):
        out = StringIO()
        call_command(
            "form_seeder",
            *args,
            stdout=out,
            stderr=StringIO(),
            **kwargs,
        )
        return out.getvalue()

    def call_fake_data_seeder_command(self, *args, **kwargs):
        out = StringIO()
        call_command(
            "fake_data_seeder",
            *args,
            stdout=out,
            stderr=StringIO(),
            **kwargs,
        )
        return out.getvalue()

    def test_call_command(self):
        forms = Forms.objects.all().delete()
        json_forms = ["Form Complaint"]
        # RUN SEED NEW FORM
        output = self.call_form_seeder_command(
            "-f", "./backend/source/forms/1.json"
        )
        if output:
            output = list(filter(lambda x: len(x), output.split("\n")))
        forms = Forms.objects.all()
        self.assertEqual(forms.count(), 1)

        for form in forms:
            self.assertIn(
                f"Form Created | {form.name} V{form.version}", output
            )
            self.assertIn(form.name, json_forms)
        # RUN UPDATE EXISTING FORM
        output = self.call_form_seeder_command(
            "-f", "./backend/source/forms/1.json"
        )
        if output:
            output = list(filter(lambda x: len(x), output.split("\n")))
        for form in forms:
            if form.version == 2:
                self.assertIn(
                    f"Form Updated | {form.name} V{form.version}", output
                )
        # RUN FAKE DATA SEEDER
        output = self.call_fake_data_seeder_command("-t", True)
        form = Forms.objects.get(name="Form Complaint")
        self.assertIn(output, form.name)

    def test_updating_form(self):
        out = StringIO()
        err = StringIO()
        first_version = "./backend/source/forms/1.json"
        call_command(
            "form_seeder",
            "--file",
            first_version,
            stdout=out,
            stderr=err,
        )
        self.assertEqual(out.getvalue(), "Form Created | Form Complaint V1\n")

        out = StringIO()
        err = StringIO()
        second_version = "./backend/source/forms/1-update-test.json"
        call_command(
            "form_seeder",
            "--file",
            second_version,
            stdout=out,
            stderr=err,
        )
        self.assertEqual(out.getvalue(), "Form Updated | Form Complaint V2\n")

    def test_seeder_raises_command_error(self):
        invalid_json_file = "./backend/source/forms/invalid.json"

        # Redirect stdout and stderr to StringIO to capture the output
        out = StringIO()
        err = StringIO()

        try:
            call_command(
                "form_seeder",
                "--file",
                invalid_json_file,
                stdout=out,
                stderr=err,
            )
            self.fail("CommandError was not raised")
        except CommandError as e:
            # Assert that the exception was raised
            self.assertEqual(
                str(e),
                f"[Errno 2] No such file or directory: '{invalid_json_file}'",
            )
        finally:
            # Print the captured output for debugging purposes
            print("stdout:", out.getvalue())
            print("stderr:", err.getvalue())



from django.test import TestCase, Client
from django.conf import settings


class CheckInitialEndpointTestCase(TestCase):
    def test_check_endpoint(self):
        client = Client()

        # Allow GET
        response = client.get("/api/gateway/check/")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json(), {"message": settings.TWILIO_ACCOUNT_SID}
        )

        # Allow GET
        response = client.post("/api/gateway/check/")
        self.assertEqual(response.status_code, 405)



from django.test import TestCase
from AkvoDjangoFormGateway.models import AkvoGatewayForm as Forms
from AkvoDjangoFormGateway.serializers import ListFormSerializer


class ListFormSerializerTestCase(TestCase):
    def setUp(self):
        self.form_attributes = {
            "id": 1,
            "name": "Form #1",
            "description": "Form #1 description",
        }
        self.form = Forms.objects.create(**self.form_attributes)
        self.serializers = ListFormSerializer(instance=self.form)

    def test_contains_expected_fields(self):
        data = self.serializers.data
        self.assertCountEqual(
            set(data.keys()), set(["id", "name", "description", "version"])
        )

    def test_version_content(self):
        data = self.serializers.data
        self.assertEqual(data["version"], self.form.version)
        self.assertEqual(str(self.form), self.form.name)

    def test_forms_api(self):
        res = self.client.get("/api/gateway/forms/1/?format=json")
        res = res.json()
        self.assertEqual(
            res,
            {
                "id": 1,
                "name": "Form #1",
                "description": "Form #1 description",
                "version": 1,
            },
        )



from rest_framework import serializers
from .models import (
    AkvoGatewayForm,
    AkvoGatewayQuestion,
    AkvoGatewayQuestionOption,
    AkvoGatewayData,
    AkvoGatewayAnswer,
)
from .constants import QuestionTypes
from .utils.functions import get_answer_value


class CheckSerializer(serializers.Serializer):
    id = serializers.IntegerField(default=1)
    check = serializers.CharField(default="OK")


class ListFormSerializer(serializers.ModelSerializer):
    class Meta:
        model = AkvoGatewayForm
        fields = ["id", "name", "description", "version"]


class ListOptionSerializer(serializers.ModelSerializer):
    class Meta:
        model = AkvoGatewayQuestionOption
        fields = ["id", "name", "order"]


class ListQuestionSerializer(serializers.ModelSerializer):
    class Meta:
        model = AkvoGatewayQuestion
        fields = ["id", "form", "order", "text"]


class ListDataSerializer(serializers.ModelSerializer):
    class Meta:
        model = AkvoGatewayData
        fields = "__all__"


class ListDataAnswerSerializer(serializers.ModelSerializer):
    value = serializers.SerializerMethodField()

    class Meta:
        model = AkvoGatewayAnswer
        fields = ["question", "value"]

    def get_value(self, instance: AkvoGatewayAnswer):
        return get_answer_value(instance)


class QuestionDefinitionSerializer(serializers.ModelSerializer):
    question_type = serializers.SerializerMethodField()
    options = serializers.SerializerMethodField()

    class Meta:
        model = AkvoGatewayQuestion
        fields = ["id", "text", "required", "question_type", "options"]

    def get_question_type(self, obj):
        return QuestionTypes.FieldStr.get(obj.type)

    def get_options(self, instance):
        options = (
            [
                options.name
                for options in instance.ag_question_question_options.all()
            ]
            if instance.ag_question_question_options.count()
            else None
        )
        return options


class TwilioSerializer(serializers.Serializer):
    Latitude = serializers.CharField(required=False)
    Longitude = serializers.CharField(required=False)
    MediaContentType0 = serializers.CharField(required=False)
    MediaUrl0 = serializers.URLField(required=False)
    SmsMessageSid = serializers.CharField(required=False)
    NumMedia = serializers.IntegerField(required=False)
    ProfileName = serializers.CharField(required=False)
    SmsSid = serializers.CharField(required=False)
    WaId = serializers.CharField(required=False)
    SmsStatus = serializers.CharField(required=False)
    Body = serializers.CharField(required=False)
    To = serializers.CharField(required=False)
    From = serializers.CharField(required=False)
    NumSegments = serializers.IntegerField(required=False)
    ReferralNumMedia = serializers.IntegerField(required=False)
    MessageSid = serializers.CharField(required=False)
    AccountSid = serializers.CharField(required=False)
    ApiVersion = serializers.CharField(required=False)
    phone = serializers.SerializerMethodField()

    def get_phone(self, obj):
        from_field = obj.get("From", "")
        if from_field.startswith("whatsapp:"):
            return from_field[10:]
        return from_field

    def validate(self, data):
        if data.get("MediaContentType0") and not data.get("MediaUrl0"):
            raise serializers.ValidationError(
                "MediaUrl0 is required when MediaContentType0 is present."
            )
        if data.get("MediaUrl0") and not data.get("MediaContentType0"):
            raise serializers.ValidationError(
                "MediaContentType0 is required when MediaUrl0 is present."
            )
        if not data.get("Latitude") and data.get("Longitude"):
            raise serializers.ValidationError(
                "Latitude is required when Longitude is present."
            )
        if data.get("Latitude") and not data.get("Longitude"):
            raise serializers.ValidationError(
                "Longitude is required when Latitude is present."
            )
        return data



# -*- coding: utf-8 -*-
import sys
import setuptools


INSTALL_PYTHON_REQUIRES = []
# We are intending to keep up to date with the supported Django versions.
# For the official support, please visit:
# https://docs.djangoproject.com/en/4.0/faq/install/#what-python-version-can-i-use-with-django
if sys.version_info[1] in [8, 9, 10, 11]:
    django_python_version_install = "Django>=4.0.4"
    INSTALL_PYTHON_REQUIRES.append(django_python_version_install)

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="AkvoDjangoFormGateway",
    author="Akvo",
    author_email="tech.consultancy@akvo.org",
    maintainer="Deden Bangkit",
    maintainer_email="deden@akvo.org",
    description=(
        "A Django library that enables seamless integration of messenger"
        " services"
    ),
    keywords="akvo, twilio, whatsapp, ssid, sms, django",
    long_description=long_description,
    long_description_content_type="text/markdown",
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    zip_safe=False,
    include_package_data=True,
    url="https://github.com/akvo/Akvo-DjangoFormGateway",
    project_urls={
        "Documentation": "https://github.com/akvo/Akvo-DjangoFormGateway",
        "Bug Reports": "https://github.com/akvo/Akvo-DjangoFormGateway/issues",
        "Source Code": "https://github.com/akvo/Akvo-DjangoFormGateway",
    },
    classifiers=[
        # https://pypi.org/classifiers/
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: Public Domain",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Framework :: Django",
        "Framework :: Django :: 4.0",
        "Framework :: Django :: 4.1",
        "Operating System :: OS Independent",
        "Topic :: Internet :: WWW/HTTP",
    ],
    python_requires=">=3.8.5",
    install_requires=[
        "setuptools>=36.2",
        "twilio>=8.2.0",
        "djangorestframework>=3.12.4",
    ]
    + INSTALL_PYTHON_REQUIRES,
    extras_require={
        "dev": ["check-manifest"],
    },
)



import random
from datetime import datetime, timedelta, time
from faker import Faker
from django.utils import timezone
from django.utils.timezone import make_aware

from AkvoDjangoFormGateway.models import (
    AkvoGatewayForm as Forms,
    AkvoGatewayData as FormData,
    AkvoGatewayAnswer as Answers,
)
from AkvoDjangoFormGateway.constants import QuestionTypes, StatusTypes


fake = Faker()


def get_answer_value(answer: Answers):
    if answer.question.type in [
        QuestionTypes.geo,
        QuestionTypes.option,
        QuestionTypes.multiple_option,
    ]:
        return answer.options
    elif answer.question.type == QuestionTypes.number:
        return answer.value
    else:
        return answer.name


def set_answer_option(data, question):
    option = None
    if question.type == QuestionTypes.option:
        option = [
            question.ag_question_question_options.order_by("?").first().name
        ]
    elif question.type == QuestionTypes.multiple_option:
        option = list(
            question.ag_question_question_options.order_by("?").values_list(
                "name", flat=True
            )[
                : fake.random_int(
                    min=2, max=question.ag_question_question_options.count()
                )
            ]
        )
    return option


def set_answer_data(data, question):
    name = None
    value = None
    option = set_answer_option(data, question)

    if question.type == QuestionTypes.geo:
        option = data.geo
    elif question.type == QuestionTypes.text:
        name = fake.sentence(nb_words=3)
    elif question.type == QuestionTypes.number:
        value = fake.random_int(min=10, max=50)
    elif question.type == QuestionTypes.photo:
        name = fake.image_url()
    elif question.type == QuestionTypes.date:
        name = fake.date_between_dates(
            date_start=timezone.datetime.now().date() - timedelta(days=90),
            date_end=timezone.datetime.now().date(),
        ).strftime("%m/%d/%Y")
    else:
        pass
    return name, value, option


def add_fake_answers(data: FormData) -> None:
    form = data.form
    number_of_answered = random.choice(
        form.ag_form_questions.values_list("id", flat=True)
    )
    for index, question in enumerate(form.ag_form_questions.all()):
        if index < number_of_answered:
            name, value, option = set_answer_data(data, question)
            Answers.objects.create(
                data=data,
                question=question,
                name=name,
                value=value,
                options=option,
            )


def seed_data(repeat: int, test: bool = False):
    for form in Forms.objects.all():
        if not test:
            print(f"\nSeeding - {form.name}")
        for i in range(repeat):
            now_date = datetime.now()
            start_date = now_date - timedelta(days=5 * 365)
            created = fake.date_between(start_date, now_date)
            created = datetime.combine(created, time.min)
            lat = fake.latitude()
            lng = fake.longitude()
            geo_value = f"{lat},{lng}"
            data = FormData.objects.create(
                form=form,
                name=fake.pystr_format(),
                phone=fake.phone_number(),
                geo=geo_value,
            )
            data.created = make_aware(created)
            data.save()
            add_fake_answers(data)
            number_of_answered = Answers.objects.filter(data=data.id).count()
            if len(form.ag_form_questions.all()) == number_of_answered:
                data.status = StatusTypes.submitted
                data.save()



import json
from datetime import datetime
from twilio.rest import Client
from .models import (
    AkvoGatewayForm as Forms,
    AkvoGatewayData as FormData,
    AkvoGatewayAnswer as Answers,
    AkvoGatewayQuestion as Questions,
)
from .constants import QuestionTypes, StatusTypes
from .utils.validation import (
    is_number,
    is_valid_string,
    is_date,
    is_valid_geolocation,
    is_valid_image,
)


class Feed:
    def __new__(cls, *args, **kwargs):
        return super().__new__(cls)

    def __init__(self):
        self.welcome = ["hi", "hello", "info"]

    def get_init_survey_session(self, text: str, form: Forms = None):
        init = False
        form_id = form.id if form else None
        if "#" in text and not form:
            info = str(text).split("#")
            form_id = info[1]
            init = len(info) == 2 and str(info[0]).lower() == "ready"
        return init, form_id

    def get_form(self, form_id: int = None, data: FormData = None) -> Forms:
        survey = None
        if data:
            survey = data.form
        if form_id:
            survey = Forms.objects.get(pk=form_id)
        return survey

    def get_last_question(self, data: FormData) -> Questions:
        if data.form.ag_form_questions.count() == data.ag_data_answer.count():
            # Return null if all questions are answered
            return None
        aws = data.ag_data_answer.all()
        ids = [aw.question.id for aw in aws]
        qs = (
            data.form.ag_form_questions.exclude(id__in=ids)
            .order_by("order")
            .first()
        )
        return qs

    def get_question(
        self, form: Forms = None, data: FormData = None
    ) -> Questions:
        qs = None
        if form:
            qs = form.ag_form_questions.all().first()
        if data and data.ag_data_answer.count():
            qs = self.get_last_question(data=data)
        return qs

    def get_draft_datapoint(self, phone: str) -> FormData:
        datapoint = (
            FormData.objects.filter(phone=phone, status=StatusTypes.draft)
            .all()
            .first()
        )
        return datapoint

    def get_list_form(self) -> str:
        forms = Forms.objects.all()
        msg = "Welcome to Akvo Survey\n\n."
        msg += "*Please select the form below:*\n"
        for f in forms:
            msg += f"- #{f.id} | {f.name}\n"
        msg += (
            "by replying to this message with the following format to start a"
            " new survey\n"
        )
        msg += "*READY#FORM_ID* (e.g *READY#1*)"
        return msg

    def validate_answer(
        self,
        text: str,
        question: Questions,
        data: FormData,
        image_type: str = None,
    ) -> None:
        # is alphanumeric by default
        is_valid = is_valid_string(input=text)
        if question.type == QuestionTypes.number:
            is_valid = is_number(input=text)
        if question.type == QuestionTypes.date:
            is_valid = is_date(input=text)
        if question.type == QuestionTypes.geo:
            is_valid = is_valid_geolocation(json_string=text)
        if question.type == QuestionTypes.photo:
            is_valid = is_valid_image(image_type=image_type)
        if question.type in [
            QuestionTypes.option,
            QuestionTypes.multiple_option,
        ]:
            lto = [str(o).lower() for o in str(text).split(",")]
            opt = [
                str(o.name).lower()
                for o in question.ag_question_question_options.all()
            ]
            count = len(set(lto).intersection(set(opt)))
            is_valid = count > 0
        return is_valid

    def insert_answer(self, text: str, question: Questions, data: FormData):
        name = None
        value = None
        options = None
        if question.type == QuestionTypes.number:
            value = text
        if question.type == QuestionTypes.geo:
            options = text
        if question.type == QuestionTypes.date:
            date_format = "%d-%m-%Y"
            dv = datetime.strptime(text, date_format)
            name = dv.strftime(date_format)
        if question.type in [
            QuestionTypes.option,
            QuestionTypes.multiple_option,
        ]:
            options = [str(t.strip()) for t in text.split(",")]
        if not name and not value and not options:
            name = text
        return Answers.objects.create(
            question=question,
            data=data,
            name=name,
            value=value,
            options=options,
        )

    def set_as_completed(self, data: FormData) -> None:
        data.status = StatusTypes.submitted
        data.save()

    def get_answer_text(
        self,
        body: str = None,
        image_url: str = None,
        lat: str = None,
        lng: str = None,
    ) -> str:
        text = body
        if image_url:
            text = image_url
        if lat and lng:
            text = json.dumps(
                [
                    lat,
                    lng,
                ]
            )
        return str(text).lower()

    def send_to_client(
        self,
        client: Client,
        from_number: str,
        body: str,
        to_number: str,
        type: str = "whatsapp",
    ):
        valid_number = len(to_number) > 8
        if valid_number:
            client.messages.create(
                from_=f"{type}:+{from_number}",
                body=body,
                to=f"{type}:+{to_number}",
            )

    def show_options(self, question: Questions) -> str:
        text = ""
        if question.type in [
            QuestionTypes.option,
            QuestionTypes.multiple_option,
        ]:
            text += "\n"
            for opt in question.ag_question_question_options.all():
                text += f"- {opt.name}\n"
            if question.type == QuestionTypes.multiple_option:
                text += (
                    "\nYou can select more than one separated by commas. *eg:"
                    " op1,op2*"
                )
        return text



# Generated by Django 4.2.1 on 2023-05-15 06:28

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="AkvoGatewayForm",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=255)),
                ("description", models.TextField(default=None, null=True)),
                ("version", models.IntegerField(default=1)),
            ],
            options={
                "db_table": "ag_form",
            },
        ),
        migrations.CreateModel(
            name="AkvoGatewayQuestion",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("order", models.BigIntegerField(default=None, null=True)),
                ("text", models.TextField()),
                (
                    "type",
                    models.IntegerField(
                        choices=[
                            (1, "Geo"),
                            (2, "Text"),
                            (3, "Number"),
                            (4, "Option"),
                            (5, "Multiple_Option"),
                            (6, "Photo"),
                            (7, "Date"),
                        ]
                    ),
                ),
                ("required", models.BooleanField(default=True, null=True)),
                (
                    "form",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="ag_form_questions",
                        to="AkvoDjangoFormGateway.akvogatewayform",
                    ),
                ),
            ],
            options={
                "db_table": "ag_question",
            },
        ),
        migrations.CreateModel(
            name="AkvoGatewayQuestionOption",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("order", models.BigIntegerField(default=None, null=True)),
                (
                    "code",
                    models.CharField(default=None, max_length=255, null=True),
                ),
                ("name", models.TextField()),
                (
                    "question",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="ag_question_question_options",
                        to="AkvoDjangoFormGateway.akvogatewayquestion",
                    ),
                ),
            ],
            options={
                "db_table": "ag_option",
            },
        ),
        migrations.CreateModel(
            name="AkvoGatewayData",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.TextField()),
                ("geo", models.JSONField(default=None, null=True)),
                ("phone", models.CharField(max_length=25)),
                ("status", models.IntegerField(default=1)),
                ("created", models.DateTimeField(auto_now_add=True)),
                ("updated", models.DateTimeField(default=None, null=True)),
                (
                    "form",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="ag_form_data",
                        to="AkvoDjangoFormGateway.akvogatewayform",
                    ),
                ),
            ],
            options={
                "db_table": "ag_data",
            },
        ),
        migrations.CreateModel(
            name="AkvoGatewayAnswer",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.TextField(default=None, null=True)),
                ("value", models.FloatField(default=None, null=True)),
                ("options", models.JSONField(default=None, null=True)),
                ("created", models.DateTimeField(auto_now_add=True)),
                ("updated", models.DateTimeField(default=None, null=True)),
                (
                    "data",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="ag_data_answer",
                        to="AkvoDjangoFormGateway.akvogatewaydata",
                    ),
                ),
                (
                    "question",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="ag_question_answer",
                        to="AkvoDjangoFormGateway.akvogatewayquestion",
                    ),
                ),
            ],
            options={
                "db_table": "ag_answer",
            },
        ),
    ]



class QuestionTypes:
    geo = 1
    text = 2
    number = 3
    option = 4
    multiple_option = 5
    photo = 6
    date = 7

    FieldStr = {
        geo: 'Geo',
        text: 'Text',
        number: 'Number',
        option: 'Option',
        multiple_option: 'Multiple_Option',
        photo: 'Photo',
        date: 'Date',
    }


class StatusTypes:
    draft = 1
    submitted = 2
